from numpy import array as _np_array, zeros as _np_zeros,exp as _np_exp 
from numpy.linalg import norm as _np_linalg_norm
from pandas import (DataFrame as _pd_DataFrame, cut as _pd_cut, concat as _pd_concat) 
from aabpl.utils.general import flatten_list
from aabpl.illustrations.illustrate_point_to_disk import illustrate_point_disk
from aabpl.illustrations.plot_pt_vars import create_plots_for_vars
from aabpl.testing.test_performance import time_func_perf
from math import pi as _math_pi
from aabpl.valid_area import disk_cell_intersection_area, disk_cell_intersection_estimate

################ aggregate_point_data_to_disks_vectorized ######################################################################################
@time_func_perf
def aggregate_point_data_to_disks_vectorized(
    grid:dict,
    pts_source:_pd_DataFrame,
    r:float,
    c:list=[],
    y:str='proj_lat',
    x:str='proj_lon',
    off_x='offset_x',
    off_y='offset_y',
    pts_target:_pd_DataFrame=None,
    row_name:str='id_y',
    col_name:str='id_x',
    cell_region_name:str='cell_region',
    sum_suffix:str=None,
    exclude_pt_itself:bool=True,
    weight_valid_area:str=None,
    plot_pt_disk:dict=None,
    silent:bool=False,
):
    """
    Aggregates Data around each point
    """
    if pts_target is None:
        pts_target = pts_source 
    # unpack grid_data 
    grid_id_to_pt_ids = grid.id_to_pt_ids
    grid_id_to_vals_xy = grid.id_to_vals_xy
    grid_id_to_sums = grid.id_to_sums
    
    grid_id_to_pt_ids_by_lvl = grid.id_to_pt_ids_by_lvl
    grid_id_to_sums_by_lvl = grid.id_to_sums_by_lvl
    grid_id_to_vals_xy_by_lvl = grid.id_to_vals_xy_by_lvl
    trynew = grid.trynew if hasattr(grid, 'trynew') else 1 
    
    sparse_grid_ids = set(grid_id_to_sums_by_lvl)
    # print("sparse_grid_ids",sparse_grid_ids)
    ids_rndm_sample = grid.ids_rndm_sample
    grid_spacing = grid.spacing
    if type(ids_rndm_sample)==bool and ids_rndm_sample:
        weight_valid_area=False # as for each point 100% of area would be valid
    else:
        grid_padding = -int(-grid_spacing//r)
        # take all cells that are part of the sampling grid
        invalid_cells = set([id for id in 
                             tuple(flatten_list([
            [(0, (int(row_id), int(col_id))) for col_id in range(min(grid.col_ids)-grid_padding, max(grid.col_ids)+grid_padding)] 
            for row_id in range(min(grid.row_ids)-grid_padding, max(grid.row_ids)+grid_padding)]))
             if not id in ids_rndm_sample])
    
    region_id_to_contained_cells = grid.search.region_id_to_contained_cells
    region_id_to_overlapped_cells = grid.search.region_id_to_overlapped_cells
    region_id_to_nested_contained_cells = grid.search.region_id_to_nested_contained_cells
    region_id_to_nested_overlapped_cells = grid.search.region_id_to_nested_overlapped_cells
    # print("region_id_to_nested_overlapped_cells[0]",region_id_to_nested_overlapped_cells[0])
    region_id_to_distinct_contained_cells = grid.search.region_id_to_distinct_contained_cells
    region_id_to_distinct_overlapped_cells = grid.search.region_id_to_distinct_overlapped_cells
    # print("region_id_to_distinct_overlapped_cells[0]",region_id_to_distinct_overlapped_cells[0])
    # cells_contained_in_all_disks = grid.search.cells_contained_in_all_disks
    # print("cells_contained_in_all_disks",cells_contained_in_all_disks)
    shared_contained_cells  = grid.search.shared_contained_cells
    shared_overlapped_cells = grid.search.shared_overlapped_cells
    
    
    # print("region_id_to_nested_overlapped_cells",region_id_to_nested_overlapped_cells)
    row_col_to_centroid = grid.row_col_to_centroid
    get_cell_centroid = grid.get_cell_centroid
    pt_id_to_xy_coords = grid.search.target.pt_id_to_xy_coords
    n_pts = len(pts_source)
    if trynew in [1]:
        shared_contained_cells_lookup = shared_contained_cells
        region_id_to_contained_cells_lookup = region_id_to_distinct_contained_cells
        shared_overlapped_cells_lookup = shared_overlapped_cells
        region_id_to_overlapped_cells_lookup = region_id_to_distinct_overlapped_cells
    elif trynew in [2]:
        shared_contained_cells_lookup = shared_contained_cells
        region_id_to_contained_cells_lookup = region_id_to_distinct_contained_cells
        shared_overlapped_cells_lookup = []
        region_id_to_overlapped_cells_lookup = region_id_to_nested_overlapped_cells
    elif trynew in [3]:
        shared_contained_cells_lookup = []
        region_id_to_contained_cells_lookup = region_id_to_nested_contained_cells
        shared_overlapped_cells_lookup = shared_overlapped_cells
        region_id_to_overlapped_cells_lookup = region_id_to_distinct_overlapped_cells
    else: 
        shared_contained_cells_lookup = []
        region_id_to_contained_cells_lookup = region_id_to_nested_contained_cells
        shared_overlapped_cells_lookup = []
        region_id_to_overlapped_cells_lookup = region_id_to_nested_overlapped_cells
    # print("///shared_contained_cells_lookup",len(shared_contained_cells_lookup), 
    #       "///shared_overlapped_cells_lookup",len(shared_overlapped_cells_lookup))
    # print("region_id_to_overlapped_cells_lookup[0]",region_id_to_overlapped_cells_lookup[0])
    # print("cells_contained_in_all_disks_incl_nested",cells_contained_in_all_disks_incl_nested)
    # initialize columns and/or reset to zero 
    
    
  
    
    # prepare plot #
    if plot_pt_disk is not None:
        if not 'pt_id' in plot_pt_disk:
            plot_pt_disk['pt_id'] = pts_source.index[int(n_pts//2)]
            plot_pt_disk['pt_id'] = sorted([(len(pt_ids), pt_ids[0] if len(pt_ids)>0 else []) for pt_ids in grid_id_to_pt_ids_by_lvl.values()])[-1][1]

    ##################### set up loop ############################
    if sum_suffix is None:
        sum_suffix = '_'+str(r)
    sum_radius_names = [(cname+sum_suffix) for cname in c]
    pts_source[sum_radius_names] = 0
     
    all_sums_cells_cntd_by_pt_cell = _np_zeros((n_pts, len(c)))
    all_sums_contained_by_pt_region = _np_zeros((n_pts, len(c)))
    all_distinct_overlap_sums = _np_zeros((n_pts, len(c)))
    all_shared_overlap_sums = _np_zeros((n_pts, len(c)))
    sums_within_disks = _np_zeros((n_pts, len(c)))
    valid_area_shares = _np_zeros(n_pts)
    valid_search_area = 2 * _math_pi * r**2
    column_dtypes = pts_target[c].dtypes
    zero_sums = _np_zeros(len(c),dtype=int) if len(c) > 1 else 0
    n_cells_pot_relevant = (-(-int(r/grid_spacing))*2+1)**2
    max_len_empty = sum(sorted([len(v) for v in grid_id_to_vals_xy.values()])[-n_cells_pot_relevant:])
    # print("n_cells_pot_relevant",n_cells_pot_relevant, "max_len_empty",max_len_empty,)
    # TODO the nest depth can be chosen according to density of the region (super cell count)
    empty_sums = _np_zeros((max_len_empty,len(c)),dtype=int)
    empty_xy_vals = _np_zeros((max_len_empty,len(c)+2),dtype=int)
    shared_overlap_sums = zero_sums
    pts_source['initial_sort'] = range(len(pts_source))
    pts_source.sort_values([row_name, col_name, cell_region_name], inplace=True)
    last_pt_row_col = (None,None)
    # last_cell_region_id = -1
    counter_new_cell = 0
    counter_new_contain_region = 0
    counter_new_overlap_region = 0
    
    # print("weight_valid_area",weight_valid_area, "trynew", trynew, 'nest_depth', grid.nest_depth)
    ############################ sum_contained_all_offset_regions #######################################################################
    if len(c) > 1:
        if weight_valid_area:
            @time_func_perf
            def sum_contained_all_offset_regions(
                    pt_row,
                    pt_col,
            ):
                """
                returns sum for cells contained in search radius for all points within cell. Additionally returns invalid area as float
                """
                cells_cntd_by_pt_cell = sparse_grid_ids.intersection(
                    [(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in shared_contained_cells_lookup])
                # invalid_area = len(invalid_cells.intersection(
                #     [(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in shared_contained_cells_lookup])) * grid_spacing**2
                # invalid_area = grid_spacing*sum(
                #     [2**(-2*lvl) for lvl, cell in invalid_cells.intersection([(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in cells_contained_in_all_disks])]
                # )
                if len(cells_cntd_by_pt_cell)>0:
                    i = 0
                    for lvl_cell in cells_cntd_by_pt_cell:
                        sums_in_cell = grid_id_to_sums_by_lvl[lvl_cell]
                        empty_sums[i:i+len(sums_in_cell)] = sums_in_cell
                        i +=len(sums_in_cell)
                    return empty_sums[:i].sum(axis=0), 0#invalid_area 
                return zero_sums, 0#invalid_area 
            #
        else:# not weight_valid_area
            @time_func_perf
            def sum_contained_all_offset_regions(
                    pt_row,
                    pt_col,
            ):
                """
                returns sum for cells contained in search radius for all points within cell. Additionally returns invalid area as float
                """
                cells_cntd_by_pt_cell = sparse_grid_ids.intersection(
                    [(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in shared_contained_cells_lookup])
                if len(cells_cntd_by_pt_cell)>0:
                    i = 0
                    for lvl_cell in cells_cntd_by_pt_cell:
                        sums_in_cell = grid_id_to_sums_by_lvl[lvl_cell]
                        empty_sums[i:i+len(sums_in_cell)] = sums_in_cell
                        i +=len(sums_in_cell)
                    return empty_sums[:i].sum(axis=0), 0 
                    # return _np_array([grid_id_to_sums_by_lvl[lvl_cell] for lvl_cell in cells_cntd_by_pt_cell]).sum(axis=0) 
                return zero_sums, 0
                #
                #
        #
    else: # len(c)==1
        if weight_valid_area:
            @time_func_perf
            def sum_contained_all_offset_regions(
                    pt_row,
                    pt_col,
            ):
                cells_cntd_by_pt_cell = sparse_grid_ids.intersection(
                    [(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in shared_contained_cells_lookup])
                # invalid_area = len(invalid_cells.intersection(
                #     [(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in shared_contained_cells_lookup])
                #     ) * grid_spacing**2
                return sum([grid_id_to_sums_by_lvl[lvl_cell] for lvl_cell in cells_cntd_by_pt_cell]), 0#invalid_area 
            #
        else:# not weight_valid_area
            @time_func_perf
            def sum_contained_all_offset_regions(
                    pt_row,
                    pt_col,
            ):
                cells_cntd_by_pt_cell = sparse_grid_ids.intersection(
                    [(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in shared_contained_cells_lookup])
                return sum([grid_id_to_sums_by_lvl[lvl_cell] for lvl_cell in cells_cntd_by_pt_cell]), 0
            #
            
        #
    ############################ sum_contained_by_offset_region #######################################################################
    if len(c) > 1:
        if weight_valid_area:
            @time_func_perf
            def sum_contained_by_offset_region(
                    pt_row,
                    pt_col,
                    cell_region_id,
            ):
                cells_contained_by_pt_region = sparse_grid_ids.intersection(
                    [(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_contained_cells_lookup[cell_region_id]])
                invalid_area = len(invalid_cells.intersection(
                    [(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_contained_cells[cell_region_id]])
                    ) * grid_spacing**2
                if len(cells_contained_by_pt_region)>0:
                    return _np_array([grid_id_to_sums_by_lvl[lvl_cell] for lvl_cell in cells_contained_by_pt_region]).sum(axis=0), invalid_area
                return zero_sums, invalid_area
            #
        else:# not weight_valid_area
            @time_func_perf
            def sum_contained_by_offset_region(
                    pt_row,
                    pt_col,
                    cell_region_id,
            ):
                cells_contained_by_pt_region = sparse_grid_ids.intersection([(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_contained_cells_lookup[cell_region_id]])
                if len(cells_contained_by_pt_region)>0:
                    return _np_array([grid_id_to_sums_by_lvl[lvl_cell] for lvl_cell in cells_contained_by_pt_region]).sum(axis=0), 0
                return zero_sums, 0
            #
        #
    else:# len(c)==1
        if weight_valid_area:
            @time_func_perf
            def sum_contained_by_offset_region(
                    pt_row,
                    pt_col,
                    cell_region_id,
            ):
                cells_contained_by_pt_region = sparse_grid_ids.intersection(
                    [(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_contained_cells_lookup[cell_region_id]])
                invalid_area = len(invalid_cells.intersection(
                    [(row+pt_row,col+pt_col) for lvl,(row,col) in region_id_to_contained_cells[cell_region_id]])
                    ) * grid_spacing**2
                return sum([grid_id_to_sums_by_lvl[lvl_cell] for lvl_cell in cells_contained_by_pt_region]), invalid_area
            #
        else:# not weight_valid_area
            @time_func_perf
            def sum_contained_by_offset_region(
                    pt_row,
                    pt_col,
                    cell_region_id,
            ):
                cells_contained_by_pt_region = sparse_grid_ids.intersection(
                    [(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_contained_cells_lookup[cell_region_id]])
                return sum([grid_id_to_sums_by_lvl[lvl_cell] for lvl_cell in cells_contained_by_pt_region]), 0
            #
        #
    #

    ############################ get_pts_overlapped_by_all #######################################################################
    @time_func_perf
    def get_pts_overlapped_by_all(
                pt_row,
                pt_col
        ):
        i = 0
        for lvl_cell in sparse_grid_ids.intersection(
                [(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in shared_overlapped_cells_lookup]):
            xy_vals_in_cell = grid_id_to_vals_xy_by_lvl[lvl_cell]
            empty_xy_vals[i:i+len(xy_vals_in_cell)] = xy_vals_in_cell
            i +=len(xy_vals_in_cell)
        return empty_xy_vals[:i]

    ############################ get_pts_overlapped_by_region #######################################################################
    if weight_valid_area:
        @time_func_perf
        def get_pts_overlapped_by_region(
                pt_row,
                pt_col,
                cell_region_id,
                n:int,
        ):  
            
            overlapped_invalid_cells = invalid_cells.intersection(
                [(row+pt_row,col+pt_col) for lvl,(row,col) in region_id_to_overlapped_cells[cell_region_id]])
            i = n
            for lvl_cell in sparse_grid_ids.intersection(
                    [(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_overlapped_cells_lookup[cell_region_id]]):
                xy_vals_in_cell = grid_id_to_vals_xy_by_lvl[lvl_cell]
                empty_xy_vals[i:i+len(xy_vals_in_cell)] = xy_vals_in_cell
                i +=len(xy_vals_in_cell)
            return empty_xy_vals[n:i], overlapped_invalid_cells
        #
    else:# not weight_valid_area
        @time_func_perf
        def get_pts_overlapped_by_region(
                pt_row,
                pt_col,
                cell_region_id,
                n:int,
        ):  
            i = n
            for lvl_cell in sparse_grid_ids.intersection(
                    [(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_overlapped_cells_lookup[cell_region_id]]):
                xy_vals_in_cell = grid_id_to_vals_xy_by_lvl[lvl_cell]
                empty_xy_vals[i:i+len(xy_vals_in_cell)] = xy_vals_in_cell
                i +=len(xy_vals_in_cell)
            return empty_xy_vals[n:i], []
            
    

    ############################ sum_overlapped_pts_in_radius #######################################################################
    if len(c) > 1:
        @time_func_perf
        def sum_overlapped_pts_in_radius(
            vals_xy_distinct_overlapped,
            pt_xycoord
        ):
            if len(vals_xy_distinct_overlapped) > 0:
                vals_in_radius = vals_xy_distinct_overlapped[:,:-2][(_np_linalg_norm( # create a mask of boolean values indicating if point are within radius 
                    vals_xy_distinct_overlapped[:,-2:] -
                    pt_xycoord, 
                axis=1) <= r)]
                
                return vals_in_radius.sum(axis=0) if len(vals_in_radius) > 0 else zero_sums
                # else no points in radius thus return vector of _np_zeros
            return zero_sums
    else:# len(c)==1
        @time_func_perf
        def sum_overlapped_pts_in_radius(
            vals_xy_distinct_overlapped,
            pt_xycoord
        ):
            if len(vals_xy_distinct_overlapped) > 0:
                return vals_xy_distinct_overlapped[:,:-2][(_np_linalg_norm( # create a mask of boolean values indicating if point are within radius 
                    vals_xy_distinct_overlapped[:,-2:] -
                    pt_xycoord, 
                axis=1) <= r)].sum(axis=0)
                # else no points in radius thus return vector of _np_zeros
            return 0

    ############################ weight_valid_area #######################################################################
    if weight_valid_area == 'precise':
        if r**2<2*grid_spacing**2:
            print("WARNING: Precise intersection method of search circle and grid cells is only implemented for search radius >= (2*grid_spacing**2)**0.5. Calculation of valid area thus might be false.")
        
        
        @time_func_perf
        def calculate_overlapped_invalid_area(
            pt_xyoffset:tuple,
            pt_row:int, 
            pt_col:int,
            invalid_overlapped_cells,
            **kwargs
        ) -> float:
            # This is slow. Either increase the speed or make a simple function that maps centroid distance to area estimate.
            
            return sum([disk_cell_intersection_area(
                    pt_xyoffset,
                    row_col=(int(row-pt_row),int(col-pt_col)), # TO-DO this
                    grid_spacing=grid_spacing,
                    r=r,
                    silent=True,
                    ) for row,col in invalid_overlapped_cells])
            
    elif weight_valid_area == 'estimate':
        
        # define here as it depends on grid_spacing / r
        @time_func_perf
        def estimate_overlapped_area_share(
            disk_center_pt_s:_np_array,      
            centroid_s:tuple=_np_array,
            logit_Q:float=1 / (0.70628102 + _np_exp(0.57266908 * (grid_spacing / r - 2))),
            logit_B:float=1 / (-0.21443453 + _np_exp(0.76899004 * (grid_spacing / r - 2))),
            r:float=r,
        ) -> _np_array:
            """
            either disk_center_pt_s or centroid_s can be more than one element not both
            returns numpy.array with share of grid cells that is overlapped by radius each element is in [0,1] or in (0,1) if cell is truly only overlapped
            """
            return 1 - 1 / (
                1.0 + logit_Q * _np_exp(
                    -logit_B * 
                        (1/r * _np_linalg_norm(disk_center_pt_s-centroid_s, axis=1) - 1)
                    )
                ) 
        
        @time_func_perf
        def calculate_overlapped_invalid_area(
            pt_xycoord,
            invalid_overlapped_cells:set,
            **kwargs
            ) -> float:
            """
            Call intersection area estimation function based on distance, radius and grid_spacing.
            Mean estimation error of 5% of cell area. Largest error for cells where only one vertex of cell lies within radius (~20%)
            """
            # This is slow. Either increase the speed or make a simple function that maps centroid distance to area estimate.
            return 0.0 if len(invalid_overlapped_cells)==0 else estimate_overlapped_area_share(
                    disk_center_pt_s=pt_xycoord,
                    centroid_s=_np_array([row_col_to_centroid.get((int(row),int(col)),get_cell_centroid(row,col)) for row,col in invalid_overlapped_cells]),
                    # centroid_s=_np_array([row_col_to_centroid[(int(row),int(col))] for row,col in invalid_overlapped_cells]),
                    ).sum() * grid_spacing ** 2
    
    else:
        
        if weight_valid_area != False and not weight_valid_area is None:
            # move to handle inputs
            print("Value for 'weight_valid_area' must be in ['precise', 'estimate', 'guess', False]. Instead",weight_valid_area,"was provided.")
        weight_valid_area = False
    #

    @time_func_perf
    def do_nothing():
        pass
    
    for (i, pt_id, pt_xycoord, pt_xyoffset, (pt_row,pt_col), contain_region_id, overlap_region_id, cell_region_id) in zip(
        range(n_pts),
        pts_source.index,
        pts_source[[x, y,]].values, 
        pts_source[[off_x, off_y,]].values, 
        pts_source[[row_name, col_name]].values,
        pts_source[cell_region_name].values // grid.search.contain_region_mult,
        pts_source[cell_region_name].values % grid.search.contain_region_mult,
        pts_source[cell_region_name].values,
        
        
        ):
        (pt_row, pt_col) = (int(pt_row), int(pt_col))
        # as pts are sorted by grid cell update only if grid cell changed
        if not (pt_row, pt_col) == last_pt_row_col:
            counter_new_cell += 1
            sums_cells_cntd_by_pt_cell, invalid_search_area_cntd_by_pt_cell = sum_contained_all_offset_regions(pt_row, pt_col)
            vals_xy_shared_overlapped = get_pts_overlapped_by_all(pt_row, pt_col)
            do_nothing()
        #
            
        if (pt_row, pt_col) != last_pt_row_col or last_contain_region_id != contain_region_id:
            counter_new_contain_region += 1
            # if cell changed or cell region changed
            (sums_contained_by_pt_region, 
             invalid_search_area_cntd_by_pt_region) = sum_contained_by_offset_region(pt_row, pt_col, cell_region_id)
            do_nothing()

        if (pt_row, pt_col) != last_pt_row_col or last_overlap_region_id != overlap_region_id:
            counter_new_overlap_region += 1
            (vals_xy_distinct_overlapped, 
             invalid_overlapped_cells) = get_pts_overlapped_by_region(pt_row, pt_col, cell_region_id, len(vals_xy_shared_overlapped))

        #
        shared_overlap_sums = sum_overlapped_pts_in_radius(vals_xy_shared_overlapped, pt_xycoord)
        distinct_overlap_sums = sum_overlapped_pts_in_radius(vals_xy_distinct_overlapped, pt_xycoord)


        # combine sums from the steps.
        # append result 
        sums_within_disks[i,:] = (
            sums_cells_cntd_by_pt_cell + 
            sums_contained_by_pt_region + 
            shared_overlap_sums + 
            distinct_overlap_sums)
        # for inspecting
        all_sums_cells_cntd_by_pt_cell[i,:] = sums_cells_cntd_by_pt_cell
        all_sums_contained_by_pt_region[i,:] = sums_contained_by_pt_region
        all_shared_overlap_sums[i,:] = shared_overlap_sums
        all_distinct_overlap_sums[i,:] = distinct_overlap_sums
        
        # calculate share of valid area
        if weight_valid_area:
            invalid_search_area_overlaps = calculate_overlapped_invalid_area(
                    pt_xyoffset=pt_xyoffset,
                    pt_xycoord=pt_xycoord,
                    pt_row=pt_row, 
                    pt_col=pt_col,
                    invalid_overlapped_cells=invalid_overlapped_cells,
                )
            valid_area_shares[i] = (
                valid_search_area - 
                invalid_search_area_cntd_by_pt_cell - 
                invalid_search_area_cntd_by_pt_region - 
                invalid_search_area_overlaps
                ) / valid_search_area
    
        # plot example pint
        if plot_pt_disk is not None and pt_id == plot_pt_disk['pt_id']:
            # cells_cntd_by_pt_cell = sparse_grid_ids.intersection([(row+pt_row,col+pt_col) for lvl,(row,col) in cells_contained_in_all_disks])
            # cells_cntd_by_pt_cell = sparse_grid_ids.intersection([(lvl, (row+pt_row,col+pt_col)) for lvl,(row,col) in shared_contained_cells_lookup])
            # print("cells_cntd_by_pt_cell",cells_cntd_by_pt_cell)
            # cells_contained_by_pt_region = sparse_grid_ids.intersection(
            #     [(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_contained_cells[cell_region_id]+cells_contained_in_all_disks])
            # print("cells_contained_by_pt_region",cells_contained_by_pt_region)
            # # print("(pt_row,pt_col)",(pt_row,pt_col))
            # # print("pt_xycoord",pt_xycoord)
            # a =  _np_array(flatten_list([
            #     grid_id_to_vals_xy_by_lvl[cell_id] for cell_id 
            #     in cells_contained_by_pt_region
            # ]))
            # print("region_id_to_overlapped_cells_incl_nested[cell_region_id]\n",region_id_to_overlapped_cells_lookup[cell_region_id])
           
            # # print("a",a)
            # # print("a.shape",a.shape)
            # # print("a[:,-2:]",a[:,-2:])
            # # print("vals_xy_distinct_overlapped",vals_xy_distinct_overlapped)
            # # for cell_id in cells_contained_by_pt_region:
            # #     print("XX",cell_id, grid_id_to_vals_xy_by_lvl[cell_id])
            
            # print("vals_xy_distinct_overlapped",vals_xy_distinct_overlapped)
            # print("linalg",_np_linalg_norm( 
            #     vals_xy_distinct_overlapped[:,-2:] -
            #     pt_xycoord, 
            # axis=1) <= r)
            vals_xy_overlapped = _np_array(list(vals_xy_distinct_overlapped)+list(vals_xy_shared_overlapped))
            
            pts_xy_in_radius = vals_xy_overlapped[:,-2:][(_np_linalg_norm( 
                vals_xy_overlapped[:,-2:] -
                pt_xycoord, 
            axis=1) <= r)]
            
            pts_xy_in_cells_overlapped_by_pt_region = vals_xy_overlapped[:,-2:][(_np_linalg_norm( 
                vals_xy_overlapped[:,-2:] -
                pt_xycoord, 
            axis=1) > r)]

            pts_xy_in_cell_contained_by_pt_region = _np_array(flatten_list([
                grid_id_to_vals_xy_by_lvl[cell_id] for cell_id 
                in sparse_grid_ids.intersection(
                    [(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in 
                     set(list(shared_contained_cells_lookup)+list(region_id_to_contained_cells_lookup[cell_region_id]))])
            ]))[:,-2:]
            
            # shared_contained_cells_lookup = []
            # region_id_to_contained_cells_lookup = region_id_to_nested_contained_cells
            # shared_overlapped_cells_lookup = []
            # region_id_to_overlapped_cells_lookup = region_id_to_nested_overlapped_cells
            illustrate_point_disk(
                grid=grid,
                pts_source=pts_source,
                pts_target=pts_target,
                r=r,
                c=c,
                x=x,
                y=y,
                shared_contained_cells=[(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in shared_contained_cells_lookup],
                shared_overlapped_cells=[(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in shared_overlapped_cells_lookup],
                distinct_contained_cells=[(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_contained_cells_lookup[cell_region_id]],
                distinct_overlapped_cells=[(lvl,(row+pt_row,col+pt_col)) for lvl,(row,col) in region_id_to_overlapped_cells_lookup[cell_region_id]],
                pts_xy_in_cell_contained_by_pt_region = pts_xy_in_cell_contained_by_pt_region,
                pts_xy_in_cells_overlapped_by_pt_region=pts_xy_in_cells_overlapped_by_pt_region,
                pts_xy_in_radius=pts_xy_in_radius,
                home_cell=(pt_row,pt_col),
                region_id=contain_region_id*grid.search.contain_region_mult+overlap_region_id,
                **plot_pt_disk,
            )
        # #

        # set id as last id for next iteration
        last_pt_row_col = (pt_row, pt_col)
        last_contain_region_id = contain_region_id
        last_overlap_region_id = overlap_region_id
    #
    pts_source[sum_radius_names] = pts_source[sum_radius_names].values + sums_within_disks
    # ensure correct dtypes
    pts_source = pts_source.astype({n:dt for n,dt, in zip(sum_radius_names, column_dtypes)})
    
    if exclude_pt_itself and grid.search.tgt_df_contains_src_df:
        # substract data from point itself unless specified otherwise
        for sum_radius_name, col_name in zip(sum_radius_names, c):
            pts_source[sum_radius_name] = pts_source[sum_radius_name].values - pts_source[col_name]
    
    if weight_valid_area:
        pts_source['valid_area_share'+sum_suffix] = valid_area_shares
        for sum_radius_name in sum_radius_names:
            pts_source[sum_radius_name] = pts_source[sum_radius_name].values / pts_source['valid_area_share'+sum_suffix].values
    
    if silent != True:
        print("Appended radius sum"+("" if len(c)<=1 else "s")+" (r="+str(r)+") for " +', '.join(["'"+cname+"' as '"+sname+"'" for (cname,sname) in zip(c, sum_radius_names)])+" to pts_source DataFrame. (Sum names can be controlled by setting sum_suffix='...')")    
        if weight_valid_area:
            print("Appended valid area share as "+'valid_area_share'+sum_suffix+" to pts_source DataFrame.")    
    # print("all_sums_cells_cntd_by_pt_cell", sum(all_sums_cells_cntd_by_pt_cell))
    # print("all_sums_contained_by_pt_region",sum(all_sums_contained_by_pt_region))#
    # print("all_shared_overlap_sums",sum(all_shared_overlap_sums))
    # print("all_distinct_overlap_sums",sum(all_distinct_overlap_sums))
    # print("FINAL RESULT---------------",sum(sums_within_disks)) 
    # print(
    #     "Share of pts in",
    #     "\n- same cell as previous:", 100-int(counter_new_cell/len(pts_source)*100),"%",
    #     "\n- same cell and containing same surrounding cells:",100 - int(counter_new_contain_region/len(pts_source)*100),"%",
    #     "\n- same cell and overlapping same surrounding cells",100 - int(counter_new_overlap_region/len(pts_source)*100),"%")
    def plot_vars(
        self = grid,
        colnames = _np_array([c, sum_radius_names]), 
        filename:str='',
        **plot_kwargs:dict,
    ):
        return create_plots_for_vars(
            grid=self,
            colnames=colnames,
            filename=filename,
            plot_kwargs=plot_kwargs,
        )

    grid.plot.vars = plot_vars

    pts_source.sort_values(['initial_sort'], inplace=True)
    pts_source.drop('initial_sort', axis=1, inplace=True)

    return pts_source[sum_radius_names]
#





"""

Rules for building C/API module with f2py2e.

Copyright 1999 -- 2011 Pearu Peterson all rights reserved.
Copyright 2011 -- present NumPy Developers.
Permission to use, modify, and distribute this software is given under the
terms of the NumPy License.

NO WARRANTY IS EXPRESSED OR IMPLIED.  USE AT YOUR OWN RISK.
"""
import copy

from .auxfuncs import (
    getfortranname, isexternal, isfunction, isfunction_wrap, isintent_in,
    isintent_out, islogicalfunction, ismoduleroutine, isscalar,
    issubroutine, issubroutine_wrap, outmess, show
)

from ._isocbind import isoc_kindmap

def var2fixfortran(vars, a, fa=None, f90mode=None):
    if fa is None:
        fa = a
    if a not in vars:
        show(vars)
        outmess('var2fixfortran: No definition for argument "%s".\n' % a)
        return ''
    if 'typespec' not in vars[a]:
        show(vars[a])
        outmess('var2fixfortran: No typespec for argument "%s".\n' % a)
        return ''
    vardef = vars[a]['typespec']
    if vardef == 'type' and 'typename' in vars[a]:
        vardef = '%s(%s)' % (vardef, vars[a]['typename'])
    selector = {}
    lk = ''
    if 'kindselector' in vars[a]:
        selector = vars[a]['kindselector']
        lk = 'kind'
    elif 'charselector' in vars[a]:
        selector = vars[a]['charselector']
        lk = 'len'
    if '*' in selector:
        if f90mode:
            if selector['*'] in ['*', ':', '(*)']:
                vardef = '%s(len=*)' % (vardef)
            else:
                vardef = '%s(%s=%s)' % (vardef, lk, selector['*'])
        else:
            if selector['*'] in ['*', ':']:
                vardef = '%s*(%s)' % (vardef, selector['*'])
            else:
                vardef = '%s*%s' % (vardef, selector['*'])
    else:
        if 'len' in selector:
            vardef = '%s(len=%s' % (vardef, selector['len'])
            if 'kind' in selector:
                vardef = '%s,kind=%s)' % (vardef, selector['kind'])
            else:
                vardef = '%s)' % (vardef)
        elif 'kind' in selector:
            vardef = '%s(kind=%s)' % (vardef, selector['kind'])

    vardef = '%s %s' % (vardef, fa)
    if 'dimension' in vars[a]:
        vardef = '%s(%s)' % (vardef, ','.join(vars[a]['dimension']))
    return vardef

def useiso_c_binding(rout):
    useisoc = False
    for key, value in rout['vars'].items():
        kind_value = value.get('kindselector', {}).get('kind')
        if kind_value in isoc_kindmap:
            return True
    return useisoc

def createfuncwrapper(rout, signature=0):
    assert isfunction(rout)

    extra_args = []
    vars = rout['vars']
    for a in rout['args']:
        v = rout['vars'][a]
        for i, d in enumerate(v.get('dimension', [])):
            if d == ':':
                dn = 'f2py_%s_d%s' % (a, i)
                dv = dict(typespec='integer', intent=['hide'])
                dv['='] = 'shape(%s, %s)' % (a, i)
                extra_args.append(dn)
                vars[dn] = dv
                v['dimension'][i] = dn
    rout['args'].extend(extra_args)
    need_interface = bool(extra_args)

    ret = ['']

    def add(line, ret=ret):
        ret[0] = '%s\n      %s' % (ret[0], line)
    name = rout['name']
    fortranname = getfortranname(rout)
    f90mode = ismoduleroutine(rout)
    newname = '%sf2pywrap' % (name)

    if newname not in vars:
        vars[newname] = vars[name]
        args = [newname] + rout['args'][1:]
    else:
        args = [newname] + rout['args']

    l_tmpl = var2fixfortran(vars, name, '@@@NAME@@@', f90mode)
    if l_tmpl[:13] == 'character*(*)':
        if f90mode:
            l_tmpl = 'character(len=10)' + l_tmpl[13:]
        else:
            l_tmpl = 'character*10' + l_tmpl[13:]
        charselect = vars[name]['charselector']
        if charselect.get('*', '') == '(*)':
            charselect['*'] = '10'

    l1 = l_tmpl.replace('@@@NAME@@@', newname)
    rl = None

    useisoc = useiso_c_binding(rout)
    sargs = ', '.join(args)
    if f90mode:
        # gh-23598 fix warning
        # Essentially, this gets called again with modules where the name of the
        # function is added to the arguments, which is not required, and removed
        sargs = sargs.replace(f"{name}, ", '')
        args = [arg for arg in args if arg != name]
        rout['args'] = args
        add('subroutine f2pywrap_%s_%s (%s)' %
            (rout['modulename'], name, sargs))
        if not signature:
            add('use %s, only : %s' % (rout['modulename'], fortranname))
        if useisoc:
            add('use iso_c_binding')
    else:
        add('subroutine f2pywrap%s (%s)' % (name, sargs))
        if useisoc:
            add('use iso_c_binding')
        if not need_interface:
            add('external %s' % (fortranname))
            rl = l_tmpl.replace('@@@NAME@@@', '') + ' ' + fortranname

    if need_interface:
        for line in rout['saved_interface'].split('\n'):
            if line.lstrip().startswith('use ') and '__user__' not in line:
                add(line)

    args = args[1:]
    dumped_args = []
    for a in args:
        if isexternal(vars[a]):
            add('external %s' % (a))
            dumped_args.append(a)
    for a in args:
        if a in dumped_args:
            continue
        if isscalar(vars[a]):
            add(var2fixfortran(vars, a, f90mode=f90mode))
            dumped_args.append(a)
    for a in args:
        if a in dumped_args:
            continue
        if isintent_in(vars[a]):
            add(var2fixfortran(vars, a, f90mode=f90mode))
            dumped_args.append(a)
    for a in args:
        if a in dumped_args:
            continue
        add(var2fixfortran(vars, a, f90mode=f90mode))

    add(l1)
    if rl is not None:
        add(rl)

    if need_interface:
        if f90mode:
            # f90 module already defines needed interface
            pass
        else:
            add('interface')
            add(rout['saved_interface'].lstrip())
            add('end interface')

    sargs = ', '.join([a for a in args if a not in extra_args])

    if not signature:
        if islogicalfunction(rout):
            add('%s = .not.(.not.%s(%s))' % (newname, fortranname, sargs))
        else:
            add('%s = %s(%s)' % (newname, fortranname, sargs))
    if f90mode:
        add('end subroutine f2pywrap_%s_%s' % (rout['modulename'], name))
    else:
        add('end')
    return ret[0]


def createsubrwrapper(rout, signature=0):
    assert issubroutine(rout)

    extra_args = []
    vars = rout['vars']
    for a in rout['args']:
        v = rout['vars'][a]
        for i, d in enumerate(v.get('dimension', [])):
            if d == ':':
                dn = 'f2py_%s_d%s' % (a, i)
                dv = dict(typespec='integer', intent=['hide'])
                dv['='] = 'shape(%s, %s)' % (a, i)
                extra_args.append(dn)
                vars[dn] = dv
                v['dimension'][i] = dn
    rout['args'].extend(extra_args)
    need_interface = bool(extra_args)

    ret = ['']

    def add(line, ret=ret):
        ret[0] = '%s\n      %s' % (ret[0], line)
    name = rout['name']
    fortranname = getfortranname(rout)
    f90mode = ismoduleroutine(rout)

    args = rout['args']

    useisoc = useiso_c_binding(rout)
    sargs = ', '.join(args)
    if f90mode:
        add('subroutine f2pywrap_%s_%s (%s)' %
            (rout['modulename'], name, sargs))
        if useisoc:
            add('use iso_c_binding')
        if not signature:
            add('use %s, only : %s' % (rout['modulename'], fortranname))
    else:
        add('subroutine f2pywrap%s (%s)' % (name, sargs))
        if useisoc:
            add('use iso_c_binding')
        if not need_interface:
            add('external %s' % (fortranname))

    if need_interface:
        for line in rout['saved_interface'].split('\n'):
            if line.lstrip().startswith('use ') and '__user__' not in line:
                add(line)

    dumped_args = []
    for a in args:
        if isexternal(vars[a]):
            add('external %s' % (a))
            dumped_args.append(a)
    for a in args:
        if a in dumped_args:
            continue
        if isscalar(vars[a]):
            add(var2fixfortran(vars, a, f90mode=f90mode))
            dumped_args.append(a)
    for a in args:
        if a in dumped_args:
            continue
        add(var2fixfortran(vars, a, f90mode=f90mode))

    if need_interface:
        if f90mode:
            # f90 module already defines needed interface
            pass
        else:
            add('interface')
            for line in rout['saved_interface'].split('\n'):
                if line.lstrip().startswith('use ') and '__user__' in line:
                    continue
                add(line)
            add('end interface')

    sargs = ', '.join([a for a in args if a not in extra_args])

    if not signature:
        add('call %s(%s)' % (fortranname, sargs))
    if f90mode:
        add('end subroutine f2pywrap_%s_%s' % (rout['modulename'], name))
    else:
        add('end')
    return ret[0]


def assubr(rout):
    if isfunction_wrap(rout):
        fortranname = getfortranname(rout)
        name = rout['name']
        outmess('\t\tCreating wrapper for Fortran function "%s"("%s")...\n' % (
            name, fortranname))
        rout = copy.copy(rout)
        fname = name
        rname = fname
        if 'result' in rout:
            rname = rout['result']
            rout['vars'][fname] = rout['vars'][rname]
        fvar = rout['vars'][fname]
        if not isintent_out(fvar):
            if 'intent' not in fvar:
                fvar['intent'] = []
            fvar['intent'].append('out')
            flag = 1
            for i in fvar['intent']:
                if i.startswith('out='):
                    flag = 0
                    break
            if flag:
                fvar['intent'].append('out=%s' % (rname))
        rout['args'][:] = [fname] + rout['args']
        return rout, createfuncwrapper(rout)
    if issubroutine_wrap(rout):
        fortranname = getfortranname(rout)
        name = rout['name']
        outmess('\t\tCreating wrapper for Fortran subroutine "%s"("%s")...\n'
                % (name, fortranname))
        rout = copy.copy(rout)
        return rout, createsubrwrapper(rout)
    return rout, ''



from math import ceil as _math_ceil
from matplotlib.pyplot import (subplots as _plt_subplots, figure as _plt_figure)
from matplotlib.patches import Circle as _plt_Circle, Rectangle as _plt_Rectangle, Polygon as _plt_Polygon
from ..utils.general import flatten_list


def plot_cell_pattern(
    contained_cells, 
    overlapped_cells, 
    all_cells,
    r:float,
    grid_spacing:float,
    nested_contained_cells=[],
    nested_overlapped_cells=[],
    region_coords:list=None, 
    add_idxs:bool=True,
    hatch='',
    **plot_kwargs,
):

    """
    Illustrate method
    region_coords: list of tuples marking coords for which to draw a circle around
    """
    # specify default plot kwargs and add defaults
    plot_kwargs = {
        'fig':None,
        'ax':None,
        's':0.8,
        'color':'#eaa',
        'figsize': (20,30),
        **plot_kwargs
    }
    figsize = plot_kwargs.pop('figsize')
    fig = plot_kwargs.pop('fig')
    ax = plot_kwargs.pop('ax')
    ###### initialize plot  ######################

    if ax is None:
        fig, ax = _plt_subplots(1,1, figsize=figsize)
    ################################################################################################################
    # add circle patches
    for region_coord in region_coords:
        ax.add_patch(
            _plt_Circle(
                xy=region_coord, radius=r,
                facecolor=('#000000'+(str(int(60/len(region_coords))) if int(60/len(region_coords))>=10 else '0'+str(int(60/len(region_coords))))),
                edgecolor='#0006', linewidth=0.25))
    colors = ['#ccc', "#9ac991ff", "#fdaba3ff"]
    # if len(nested_contained_cells)+len(nested_overlapped_cells)>0:
    #     overlapped_cells = []
    [[((lvl,(row, col)), color) for lvl,(row, col) in cells] for cells,color in zip(
        [all_cells, contained_cells, overlapped_cells],
        colors)]
    for (lvl,(row, col)), color in flatten_list(
        [[((lvl,(row, col)), color) for lvl,(row, col) in cells] for cells,color in zip(
        [all_cells, contained_cells, overlapped_cells],
        colors)]):
        lvl=lvl
        ax.add_patch(_plt_Rectangle(
            xy = (float(col)-2**-lvl, float(row)-2**-lvl), 
            width=2**(1-lvl), height=2**(1-lvl), 
            linewidth=.7, facecolor=color, edgecolor=color, alpha=1, 
            **({} if hatch =='' else {'hatch':hatch})
        ))
        
        if add_idxs and color == colors[0]:
            ax.annotate(text=str(row)+","+str(col), xy=(col,row),horizontalalignment='center')
    
    colors = ["#1d802a", "#a51414"]
    for (lvl,(row, col)), color in flatten_list(
        [[((lvl,(row, col)), color) for lvl, (row, col)  in cells] for cells,color in zip(
        [nested_contained_cells, nested_overlapped_cells],
        colors)]):
        xy = (float(col)-2**(-lvl-1), float(row)-2**(-lvl-1))
        ax.add_patch(_plt_Rectangle(
            xy = xy, 
            width=2**(-lvl), height=2**(-lvl), 
            linewidth=.7, facecolor=color, edgecolor=color, alpha=0.3, 
            **({} if hatch =='' else {'hatch':hatch})
        ))
    ax.add_patch(_plt_Rectangle(
            xy = (-.5,-.5), 
            width=1, height=1, 
            linewidth=.7, facecolor='white', edgecolor='black', alpha=0.3, 
            **({} if hatch =='' else {'hatch':hatch})
        ))
    
    
    # add region polygon
    ax.add_patch(
        _plt_Polygon(
            xy=region_coords,
            facecolor="#FFFFFF",
            edgecolor='#000', linewidth=0.25))
        
    ratio = r/grid_spacing
    cell_steps_max = _math_ceil(ratio+1.5)

    ax.set_xlim((-cell_steps_max,+cell_steps_max))
    ax.set_ylim((-cell_steps_max,+cell_steps_max))
    ax.set_aspect('equal', adjustable='box')
    #
#



# install package into your environment through your console via
# pip install ABRSQOL
# or install it from this script:
import subprocess, sys
subprocess.check_call([sys.executable, "-m", "pip", "install", 'aabpl', "--upgrade"])

### set up working directory and folders
import os
# Create output folders if they don't exist
working_directory = "P:/Algorithm/PL_python/AABPL-toolkit-python" # specify your folder path here C:/User/YourName/YourFolder
output_data_folder = os.path.join(working_directory, "output_data/")
output_gis_folder = os.path.join(working_directory, "output_gis/")
output_maps_folder = os.path.join(working_directory, "output_maps/")
temp_folder = os.path.join(working_directory, "temp")
os.makedirs(output_data_folder, exist_ok=True)
os.makedirs(output_gis_folder, exist_ok=True)
os.makedirs(output_maps_folder, exist_ok=True)
os.makedirs(temp_folder, exist_ok=True)

### Import packages
from pandas import read_csv
from aabpl.main import radius_search, detect_cluster_pts, detect_cluster_cells
from aabpl.valid_area import infer_sample_area_from_pts

path_to_your_csv = 'input_data/hist_New_York.txt'
crs_of_your_csv =  "EPSG:4326"
pts = read_csv(path_to_your_csv, sep=",", header=None)
pts.columns = ["eid", "employment", "industry", "lat","lon","moved"]

sample_area = infer_sample_area_from_pts(
    pts=pts,
    x='lon',
    y='lat',
    hull_type='concave',
    concavity=0.2,
    buffer=750,
    tolerance=750,
    plot_sample_area=True,
)

grid = detect_cluster_cells(
    pts=pts,
    crs=crs_of_your_csv,
    r=750,
    c='employment', # Name of the column or list of columns for which values shall be aggregated within search radius 
    exclude_pt_itself=True, # False
    sample_area=sample_area, # 'concave', 'convex', 'buffer', 'bounding_box', 'grid' or None
    min_pts_to_sample_cell=0, # 1, 10
    weight_valid_area=None, # 'estimate', 'precise'
    k_th_percentile=99.5,
    n_random_points=100000,
    random_seed=0,
    queen_contingency=1, # 0, 2
    centroid_dist_threshold=2500,
    border_dist_threshold=1000,
    min_cluster_share_after_contingency=0.05,
    make_convex=True, # False
)

## Save DataFrames with radius sums and clusters
# Using all the save options below is most likely excessive. 
# saving the shapefile for save_cell_clusters and save_sparse_grid is most
# likely sufficient

# save files as needed
# save only only clusters including their geometry, aggregate values, area and id
df_clusters = grid.save_cell_clusters(filename=output_gis_folder+'clusters', file_format='shp')
df_clusters = grid.save_cell_clusters(filename=output_data_folder+'clusters', file_format='csv')

# save sparse grid including cells only those cells that at least contain one point
df_sparse_grid = grid.save_sparse_grid(filename=output_gis_folder+'sparse_grid', file_format='shp')
df_sparse_grid = grid.save_sparse_grid(filename=output_data_folder+'sparse_grid', file_format='csv')

# save full grid including cells that have no points in them (through many empty cells this will occuppy unecessary disk space)
# df_full_grid = grid.save_full_grid(filename=output_gis_folder+'full_grid', file_format='shp')
# df_full_grid = grid.save_full_grid(filename=output_data_folder+'full_grid', file_format='csv')

pts.to_csv(output_data_folder+'pts_df_w_clusters.csv')

# CREATE PLOTS
fig = grid.plot.clusters(output_maps_folder+'clusters_employment_750m_995th')
fig = grid.plot.vars(filename=output_maps_folder+'employment_vars')
fig = grid.plot.cluster_pts(filename=output_maps_folder+'employment_cluster_pts')
fig = grid.plot.rand_dist(filename=output_maps_folder+'rand_dist_employment')

print("Successfully executed Example.py")





from numpy import (
    array as _np_array,
    append as _np_append,
    invert as _np_invert,
    zeros as _np_zeros,
    min as _np_min, 
    max as _np_max, 
    logical_or as _np_logical_or, 
    all as _np_all, 
)
from numpy.linalg import norm as _np_linalg_norm
from pandas import (
    DataFrame as _pd_DataFrame, 
    cut as _pd_cut,
) 

def create_nested_cell_structure(
    cell_dict:dict,
    nest_levels_remaining:int=5,
    max_pts_per_cell:int=10,
):
    """
    input cell contains at least one pt
    """
    

    pts_lat_lon = cell_dict['pts_lat_lon']
    # shrink cell bounds to minimal bounding box containing all its points
    if len(pts_lat_lon)==0:
        print(nest_levels_remaining, "NOLEN",cell_dict)
    lat_min, lon_min, lat_max, lon_max = pts_lat_lon[:,0].min(), pts_lat_lon[:,1].min(), pts_lat_lon[:,0].max(), pts_lat_lon[:,1].max()
    cell_dict['bounds'] = ((lat_min, lon_min), (lat_max, lon_max))

    # early return if
    if ( 
        # less pts then max_pts_per_cell specifies 
        len(cell_dict['pt_ids']) <= max_pts_per_cell or
        # maximum nest levels reached
        nest_levels_remaining == 0 or
        # if all pts are exactly in the same location  
        (lat_min==lat_max and lon_min==lon_max)
        ):
        # return unchanged (except for tightend bounds)
        return cell_dict 
    
    
    pt_ids = cell_dict['pt_ids']
    pts_vals = cell_dict['pts_vals']
    
    # find largest extension in dimensions and split in half s.t.
    # split along north-south border
    # split along west - east border 
    threshold_dimension = int(lat_max-lat_min < lon_max-lon_min)
    # find value that splits the pts in half
    threshold_upper_value = sorted(pts_lat_lon[:,threshold_dimension])[int(len(pts_lat_lon)/2)]
    if not (pts_lat_lon[:,threshold_dimension]<threshold_upper_value).any():
        # handle 
        threshold_lower_value = threshold_upper_value
        threshold_upper_value = _np_min(pts_lat_lon[:,threshold_dimension][pts_lat_lon[:,threshold_dimension]>threshold_upper_value])
    else: 
        threshold_lower_value = _np_max(pts_lat_lon[:,threshold_dimension][pts_lat_lon[:,threshold_dimension]<threshold_upper_value])
    # use this itermediate value to ensure to split pts even if threshold value is repeated accross pts
    threshold_value = (threshold_upper_value+threshold_lower_value)/2
  
    
    pt_is_below_threshold = pts_lat_lon[:,threshold_dimension] <= threshold_value

    
    bounds_below_threshold = (
        (lat_min, lon_min), 
        ((lat_max,threshold_lower_value)[threshold_dimension], (threshold_lower_value,lon_max)[threshold_dimension])
        )
    bounds_above_threshold = (
        ((lat_min,threshold_upper_value)[threshold_dimension], (threshold_upper_value, lon_min)[threshold_dimension]), 
        (lat_max, lon_max)
        )


    # append list of quadrants
    cell_dict['quadrants'] = [create_nested_cell_structure(subcell,nest_levels_remaining-1,max_pts_per_cell) for subcell in [
        {
            'sums': pts_vals[pt_is_below_threshold].sum(axis=0),
            'pt_ids': pt_ids[pt_is_below_threshold],
            'pts_vals': pts_vals[pt_is_below_threshold],
            'pts_lat_lon': pts_lat_lon[pt_is_below_threshold],
            'bounds':bounds_below_threshold,
        },
        {
            'sums': pts_vals[_np_invert(pt_is_below_threshold)].sum(axis=0),
            'pt_ids': pt_ids[_np_invert(pt_is_below_threshold)],
            'pts_vals': pts_vals[_np_invert(pt_is_below_threshold)],
            'pts_lat_lon': pts_lat_lon[_np_invert(pt_is_below_threshold)],
            'bounds':bounds_above_threshold,
        }
    ]]
    
    # remove keys that are not necesssary any longer
    del cell_dict['pt_ids']
    del cell_dict['pts_vals']
    del cell_dict['pts_lat_lon']

    return cell_dict
#

def create_nested_cell_structure_quadrants(
    cell_dict:dict,
    nest_levels_remaining:int=5,
    max_pts_per_cell:int=10,
):
    """
    TODO this is not currently used. 
    TODO it does not recursively call itself but another function. is that right? 
    input cell contains at least one pt
    """
    

    pts_lat_lon = cell_dict['pts_lat_lon']
    # shrink cell bounds to minimal bounding box containing all its points

    lat_min, lon_min, lat_max, lon_max = pts_lat_lon[:,0].min(), pts_lat_lon[:,1].min(), pts_lat_lon[:,0].max(), pts_lat_lon[:,1].max()
    cell_dict['bounds'] = ((lat_min, lon_min), (lat_max, lon_max))

    # early return if
    if ( 
        # less pts then max_pts_per_cell specifies 
        len(cell_dict['pt_ids']) <= max_pts_per_cell or
        # maximum nest levels reached
        nest_levels_remaining == 0 or
        # if all pts are exactly in the same location  
        (lat_min==lat_max and lon_min==lon_max)
        ):
        # return unchanged (except for tightend bounds)
        return cell_dict 
    
    
    pt_ids = cell_dict['pt_ids']
    pts_vals = cell_dict['pts_vals']
    
    quadrants_bounds = [
        # quadrant 1: north west
        [( (lat_min+lat_max)/2,             lon_min ), (             lat_max, (lon_min+lon_max)/2 )],
        # quadrant 2: north east
        [( (lat_min+lat_max)/2, (lon_min+lon_max)/2 ), (             lat_max,             lon_max )],
        # quadrant 3: south west
        [(             lat_min,             lon_min ), ( (lat_min+lat_max)/2, (lon_min+lon_max)/2 )],
        # quadrant 4: south east
        [(             lat_min, (lon_min+lon_max)/2 ), ( (lat_min+lat_max)/2,             lon_max )],
        ]
    
    quadrants_pt_is_in = [
            # ensure that that top row includes pts on its northern boundary
            (q_lat_min <= pts_lat_lon[:,0]) * (pts_lat_lon[:,0] < q_lat_max+int(q_nr+1 in (1,2))) * 
            # ensure that that right col includes pts on its eastern boundary
            (q_lon_min <= pts_lat_lon[:,1]) * (pts_lat_lon[:,1] < q_lon_max+int(q_nr+1 in (2,4)))   
          for q_nr, ((q_lat_min, q_lon_min), (q_lat_max, q_lon_max)) in enumerate(quadrants_bounds)]

    non_empty_quadrant_dict_list = [{
            'sums': pts_vals[pt_is_in_quad].sum(axis=0),
            'pt_ids': pt_ids[pt_is_in_quad],
            'pts_vals': pts_vals[pt_is_in_quad],
            'pts_lat_lon': pts_lat_lon[pt_is_in_quad],
            'bounds':quad_bounds,
        } for pt_is_in_quad, quad_bounds in  zip(quadrants_pt_is_in, quadrants_bounds) if any(pt_is_in_quad)]
    if len(non_empty_quadrant_dict_list)==0:
        print("ERROR!")
    # append list of quadrants
    cell_dict['quadrants'] = [create_nested_cell_structure(quadrant,nest_levels_remaining-1,max_pts_per_cell) for quadrant in non_empty_quadrant_dict_list]
    
    # remove keys that are not necesssary any longer
    del cell_dict['pt_ids']
    del cell_dict['pts_vals']
    del cell_dict['pts_lat_lon']

    return cell_dict
#

def aggregate_point_data_to_nested_cells(
    grid:dict,
    pts_df:_pd_DataFrame,
    sum_names:list=['employment'],
    y_coord_name:str='lat',
    x_coord_name:str='lon',
    cell_name:str='cell_id',
    row_name:str='id_y',
    col_name:str='id_x',
    max_nest_levels:int=5,
    max_pts_per_cell:int=10,
    silent = False,
) -> _pd_DataFrame:
    """
    Inputs:
    Modifies input pandas.DataFrame grid and pts_df: 
    - sorts by 1) y coordinate and 2) by x coordinate
    - 
    Returns 
    gridcell_id_name: name to be appended in pts_df to indicate gridcell. If False then information will not be stored in pts_df 
    """
    # TO Do this might be significantly faster when looping through pts_df instead of through cells
    pts_df.sort_values([y_coord_name, x_coord_name], inplace=True)

    # . 
    row_ids = grid.row_ids
    col_ids = grid.col_ids
    # get vectors of row columns boundary values
    y_steps=grid.y_steps
    x_steps=grid.x_steps
    # store len and digits for index
    id_y_mult = grid.id_y_mult
    len_pts_df = len(pts_df)

    # initialize dicts for later lookups

    grid.pt_id_to_row_col = {}
    nested_grid = {}
    grid_id_to_bounds = grid.id_to_bounds
    if not silent:
        print(
            'Aggregate Data from '+str(len_pts_df)+' points'+
            ' into '+str(len(y_steps))+'x'+str(len(x_steps))+
            '='+str(len(y_steps)*len(x_steps))+' cells for '+str(len(sum_names))+' indicator(s).' 
            )

   
    # to do change to cut
    # for each row select relevant points, then refine selection with columns to obtain cells
    pts_df[row_name]=_pd_cut(
        x=pts_df[y_coord_name],
        bins=y_steps,labels=row_ids,include_lowest=True).astype(int)
    
    pts_df[col_name]=_pd_cut(
        x=pts_df[x_coord_name],
        bins=x_steps,labels=col_ids,include_lowest=True).astype(int)
    
    pts_df[cell_name] = id_y_mult * pts_df[row_name] + pts_df[col_name]

    for pt_id, pt_cell_id, pt_row_id, pt_col_id, pt_vals, pt_lat_lon in zip(
        pts_df.index, 
        pts_df[cell_name], 
        pts_df[row_name],
        pts_df[col_name], 
        pts_df[sum_names].values,
        pts_df[[y_coord_name,x_coord_name]].values
        ):
        # grid.id_to_pt_ids[pt_cell_id] = _np_append(grid.id_to_pt_ids[pt_cell_id], pt_id)
        # grid.id_to_sums[pt_cell_id]=grid.id_to_sums[pt_cell_id]+pt_vals
        grid.pt_id_to_row_col[pt_id] = (pt_row_id, pt_col_id)
        
        # grid.pt_id_to_lat_lon[pt_id] = (pt_row_id, pt_col_id)

        # nested TODO use classes here.
        if not pt_cell_id in nested_grid:
            # initialize cell dict
            nested_grid[pt_cell_id] = {
            'sums':pt_vals,
            'pt_ids':[pt_id], 
            'pts_vals':[pt_vals],
            'pts_lat_lon':[pt_lat_lon],
            'bounds': grid_id_to_bounds[pt_cell_id],
        }
        else:
            cell_dict = nested_grid[pt_cell_id]
            cell_dict['sums'] = cell_dict['sums'] + pt_vals 
            cell_dict['pt_ids'].append(pt_id) 
            cell_dict['pts_vals'].append(pt_vals) 
            cell_dict['pts_lat_lon'].append(pt_lat_lon) 
        #
    #
    
   
    # create neste structure of grid cell 
    # set parameters
    for grid_id,cell_dict in nested_grid.items():

        # convert to arrays
        cell_dict['pt_ids'] = _np_array(cell_dict['pt_ids'])
        cell_dict['pts_vals'] = _np_array(cell_dict['pts_vals'])
        cell_dict['pts_lat_lon'] = _np_array(cell_dict['pts_lat_lon'])

        nested_grid[grid_id] = create_nested_cell_structure(
            cell_dict=cell_dict,
            nest_levels_remaining=max_nest_levels,
            max_pts_per_cell=max_pts_per_cell,
            )
        
    grid.nested = nested_grid

    if not silent:
        print(
            'Points assigned to grid cell:'+
            str(len(pts_df.index) - _np_logical_or(pts_df[col_name]==-1, pts_df[row_name]==-1).sum())+
            '/'+str(len(pts_df.index))
        )
    return 
#

######################## PERFORM_SERACH #####################


def nested_distance_checks_directional(
    cell_dict:dict,
        pt_lat:float,
        pt_lon:float,
        pt_is_north_of_cell:bool,
        pt_is_east_of_cell:bool,
        r:float=0.0074,
        zeros:_np_array=_np_zeros(1,dtype=int),
):
    """
    
    """
    # cell_dict['bounds]
    (cell_lat_min, cell_lon_min), (cell_lat_max, cell_lon_max) = cell_dict['bounds']
    (cell_lat_min, cell_lat_max)[int(pt_is_north_of_cell)]
    

    # else point is in between cell bounds at leas in one dimension thus two points have to be checked
    farthest_vertex_in_disk = (
            (pt_lat-(cell_lat_min if pt_is_north_of_cell else cell_lat_max))**2 + 
            (pt_lon-(cell_lon_min if pt_is_east_of_cell  else cell_lon_max))**2
            )**.5 <= r

    # cell contained in disk if
    if farthest_vertex_in_disk:
        return cell_dict['sums']
    
    closest_vertex_in_disk = (
            (pt_lat-(cell_lat_max if pt_is_north_of_cell else cell_lat_min))**2 + 
            (pt_lon-(cell_lon_max if pt_is_east_of_cell  else cell_lon_min))**2
            )**.5 <= r

    # overlapped if
    if closest_vertex_in_disk:
        # if there are quadrants within cell search downwwards
        if 'quadrants' in cell_dict:
            return _np_array([nested_distance_checks_directional(
                    cell_dict=quadrant,
                    pt_lat=pt_lat,
                    pt_lon=pt_lon,
                    pt_is_north_of_cell=pt_is_north_of_cell,
                    pt_is_east_of_cell=pt_is_east_of_cell,
                    r=r,
                    zeros=zeros,
            ) for quadrant in cell_dict['quadrants']]).sum(axis=0)
        
        # else: There are not quadrants - thus deepest level has been reached   
        # check which pts are within disk and retrieve their vals and create sum
        return cell_dict['pts_vals'][_np_linalg_norm(cell_dict['pts_lat_lon']-_np_array([pt_lat,pt_lon]),axis=1)<=r].sum(axis=0) 
    
    # cell does not intersects disk
    return zeros
#

def nested_distance_checks(
        cell_dict:dict,
            pt_lat:float,
            pt_lon:float,
            r:float=0.0074,
            zeros:_np_array=_np_zeros(1,dtype=int),
):
    """
    
    """
    # directly evalute pairwise distance if at lowest nest level to avoid unnessary overhead
    if 'pts_vals' in cell_dict:
        return cell_dict['pts_vals'][_np_linalg_norm(cell_dict['pts_lat_lon']-_np_array([pt_lat,pt_lon]),axis=1)<=r].sum(axis=0)


    # unpack for better readability
    (cell_lat_min, cell_lon_min), (cell_lat_max, cell_lon_max) = cell_dict['bounds']
        
    # check if pt not within cell lat or lon 'range'
    if (
        # pt is not (btwn_lat_min_max)
        not ((pt_lat <= cell_lat_min) != (pt_lat < cell_lat_max)) and 
        # pt is not (btwn_lon_min_max)
        not ((pt_lon <= cell_lon_min) != (pt_lon < cell_lon_max))
    ):
        # pt is not between max and min lat / lon values of cell
        # thus it is clear which vertex (north-west, north-east,south-east,south-west) is closest and which farthest from pt
        # the same vertex will also be closest for any possibly nested subcells
        
        return nested_distance_checks_directional(
                cell_dict=cell_dict,
                    pt_lat=pt_lat,
                    pt_lon=pt_lon,
                    # pt is north of cell lower if true else its south (as beeing in between is excluded)
                    pt_is_north_of_cell=pt_lat>cell_lat_min,
                    # pt is east of cell lower if true else its west (as beeing in between is excluded)
                    pt_is_east_of_cell=pt_lon>cell_lon_min,
                    r=r,
                    zeros=zeros,
            )
    
    return _np_array([nested_distance_checks(
                cell_dict=quadrant,
                    pt_lat=pt_lat,
                    pt_lon=pt_lon,
                    r=r,
                    zeros=zeros,
            ) for quadrant in cell_dict['quadrants']]).sum(axis=0) 


    # AVOID TROUGH EARLY RETURN 


    # else point is in between cell bounds at leas in one dimension thus two points have to be checked
    vertex_1_in_disk = ((pt_lat-cell_lat_min)**2+(pt_lon-cell_lon_min)**2)**.5 <= r
    vertex_2_in_disk = ((pt_lat-cell_lat_max)**2+(pt_lon-cell_lon_min)**2)**.5 <= r
    vertex_3_in_disk = ((pt_lat-cell_lat_min)**2+(pt_lon-cell_lon_max)**2)**.5 <= r
    vertex_4_in_disk = ((pt_lat-cell_lat_max)**2+(pt_lon-cell_lon_max)**2)**.5 <= r

    # if cell contained in disk return cell sum
    if _np_all([vertex_1_in_disk, vertex_2_in_disk, vertex_3_in_disk, vertex_4_in_disk]):
        return cell_dict['sums']
    
    # if overlapped
    if any([vertex_1_in_disk, vertex_2_in_disk, vertex_3_in_disk, vertex_4_in_disk]):
        
        # if no further sub quadrants check distance to pts
        # handle case where disk overlaps but no pts within it are within disk
        return _np_array([nested_distance_checks(
                cell_dict=quadrant,
                    pt_lat=pt_lat,
                    pt_lon=pt_lon,
                    r=r,
                    zeros=zeros,
            ) for quadrant in cell_dict['quadrants']]).sum(axis=0) 
    # cell does not intersects disk
    return zeros
#


def aggreagate_point_data_to_disks_vectorized_nested(
    grid:dict,
    pts_df:_pd_DataFrame,
    r:float=0.0075,
    sum_names:list=['employment'],
    y_coord_name:str='lat',
    x_coord_name:str='lon',
    row_name:str='id_y',
    col_name:str='id_x',
    cell_name:str='cell_id',
    sum_suffix:str='_750m',
    exclude_pt_itself:bool=True,
    reset_sum_cols_to_zero:bool=True,
    silent = False,
):
    """
    
    """
    # unpack grid_data 
    grid_spacing = grid.spacing
    # grid_id_to_sums = grid.id_to_sums
    pt_id_to_grid_xy = grid.pt_id_to_row_col
    region_id_to_contained_cells = grid.search.region_id_to_contained_cells
    region_id_to_overlapped_cells = grid.search.region_id_to_overlapped_cells
    nested_grid = grid.nested
    id_y_mult=grid.id_y_mult
    cells_contained_in_all_disks =grid.search.cells_contained_in_all_disks

    # prepare dicts for fast lookup of values for point ids
 
    # initialize columns and/or reset to zero unless specified differently
    sum_radius_names = [(cname+sum_suffix) for cname in sum_names]
    pts_df[[cname for cname in sum_radius_names if reset_sum_cols_to_zero or not cname in pts_df.columns]] = 0
    
    # sort according to grid cell
    # also sort according to grid cell region!
    pts_df.sort_values([row_name, col_name, cell_region_name], inplace=True)
    store_index = pts_df.index 
    pts_df.reset_index(inplace=True,drop=True)
    
  
    sums_within_disks = []
    last_grid_id = -1
    last_cell_region_id = -1
    zero_sums = _np_zeros(len(sum_names),dtype=int)
    sparse_grid_ids = set(pts_df[cell_name])

    for pt_id, a_pt_ycoord, a_pt_xcoord, cell_id, cell_region_id in zip(
        store_index,
        pts_df[y_coord_name], 
        pts_df[x_coord_name],
        pts_df[cell_name],
        pts_df[cell_region_name],
        ):
        (grid_id_y, grid_id_x) = pt_id_to_grid_xy[pt_id]
        
        # as pts are sorted by grid cell update only if grid cell changed
        if not (grid_id_y, grid_id_x) == last_grid_id:
            
            contained_cell_ids = [cell_id for cell_id in (id_y_mult*(cells_contained_in_all_disks[:,0]+(grid_id_y))+(
                cells_contained_in_all_disks[:,1]+grid_id_x)) if cell_id in nested_grid] 
       
            contained_cells_sums = (_np_array([nested_grid[g_id]['sums'] for g_id in contained_cell_ids]).sum(axis=0) 
                                      if len(contained_cell_ids)>0 else zero_sums)
        #
            
        if not (grid_id_y, grid_id_x) == last_grid_id or last_cell_region_id != cell_region_id:
            # if cell changed or cell region changed
            cell_ids_full_in_cell_region = [cell_id for cell_id in (id_y_mult*(region_id_to_contained_cells[cell_region_id][:,0]+(grid_id_y))+(
                region_id_to_contained_cells[cell_region_id][:,1]+grid_id_x)) if cell_id in nested_grid]
                           
            contained_cells_sums_region = (_np_array([nested_grid[g_id]['sums'] for g_id in cell_ids_full_in_cell_region if g_id in nested_grid]).sum(axis=0) 
                                             if len(cell_ids_full_in_cell_region)>0 else zero_sums)

            cell_ids_maybe_in = [cell_id for cell_id in (id_y_mult*(region_id_to_overlapped_cells[cell_region_id][:,0]+(grid_id_y))+(
                region_id_to_overlapped_cells[cell_region_id][:,1]+grid_id_x)) if cell_id in nested_grid]
        #

        sums_from_overlapping_cells = [nested_distance_checks(
                cell_dict = nested_grid[cell_id],
                pt_lat = a_pt_ycoord,
                pt_lon = a_pt_xcoord,
                zeros = zero_sums
        ) for cell_id in cell_ids_maybe_in]

        overlapping_cells_sums = _np_array(sums_from_overlapping_cells).sum(axis=0) if len(sums_from_overlapping_cells)>0 else zero_sums

        sums_full_disk = contained_cells_sums + contained_cells_sums_region + overlapping_cells_sums
        
        # append result 
        sums_within_disks.append(sums_full_disk)
        
        # set id as last id for next iteration
        last_grid_id=(grid_id_y, grid_id_x)
        last_cell_region_id = cell_region_id
    #

    pts_df[sum_radius_names] = pts_df[sum_radius_names].values + sums_within_disks
            
    if exclude_pt_itself:
        # substract data from point itself unless specified otherwise
        pts_df[sum_radius_names] = pts_df[sum_radius_names].values-pts_df[sum_names]
    
    # restore index 
    pts_df.index = store_index

    return 
#


import sys
from typing import Any

import numpy as np
import numpy.typing as npt

if sys.version_info >= (3, 11):
    from typing import assert_type
else:
    from typing_extensions import assert_type

i8: np.int64
f8: np.float64

AR_b: npt.NDArray[np.bool]
AR_i8: npt.NDArray[np.int64]
AR_f8: npt.NDArray[np.float64]

AR_LIKE_f8: list[float]

assert_type(np.take_along_axis(AR_f8, AR_i8, axis=1), npt.NDArray[np.float64])
assert_type(np.take_along_axis(f8, AR_i8, axis=None), npt.NDArray[np.float64])

assert_type(np.put_along_axis(AR_f8, AR_i8, "1.0", axis=1), None)

assert_type(np.expand_dims(AR_i8, 2), npt.NDArray[np.int64])
assert_type(np.expand_dims(AR_LIKE_f8, 2), npt.NDArray[Any])

assert_type(np.column_stack([AR_i8]), npt.NDArray[np.int64])
assert_type(np.column_stack([AR_LIKE_f8]), npt.NDArray[Any])

assert_type(np.dstack([AR_i8]), npt.NDArray[np.int64])
assert_type(np.dstack([AR_LIKE_f8]), npt.NDArray[Any])

assert_type(np.array_split(AR_i8, [3, 5, 6, 10]), list[npt.NDArray[np.int64]])
assert_type(np.array_split(AR_LIKE_f8, [3, 5, 6, 10]), list[npt.NDArray[Any]])

assert_type(np.split(AR_i8, [3, 5, 6, 10]), list[npt.NDArray[np.int64]])
assert_type(np.split(AR_LIKE_f8, [3, 5, 6, 10]), list[npt.NDArray[Any]])

assert_type(np.hsplit(AR_i8, [3, 5, 6, 10]), list[npt.NDArray[np.int64]])
assert_type(np.hsplit(AR_LIKE_f8, [3, 5, 6, 10]), list[npt.NDArray[Any]])

assert_type(np.vsplit(AR_i8, [3, 5, 6, 10]), list[npt.NDArray[np.int64]])
assert_type(np.vsplit(AR_LIKE_f8, [3, 5, 6, 10]), list[npt.NDArray[Any]])

assert_type(np.dsplit(AR_i8, [3, 5, 6, 10]), list[npt.NDArray[np.int64]])
assert_type(np.dsplit(AR_LIKE_f8, [3, 5, 6, 10]), list[npt.NDArray[Any]])

assert_type(np.kron(AR_b, AR_b), npt.NDArray[np.bool])
assert_type(np.kron(AR_b, AR_i8), npt.NDArray[np.signedinteger[Any]])
assert_type(np.kron(AR_f8, AR_f8), npt.NDArray[np.floating[Any]])

assert_type(np.tile(AR_i8, 5), npt.NDArray[np.int64])
assert_type(np.tile(AR_LIKE_f8, [2, 2]), npt.NDArray[Any])



from numpy import (
    array as _np_array,
    zeros as _np_zeros,
    linspace as _np_linspace
)
from pandas import (DataFrame as _pd_DataFrame, cut as _pd_cut, concat as _pd_concat) 
from aabpl.utils.general import find_column_name,arr_to_tpls
from aabpl.testing.test_performance import time_func_perf
# from aabpl.doc.docstrings import fixdocstring

################ assign_points_to_cells ######################################################################################
# @fixdocstring
@time_func_perf
def assign_points_to_cells(
    grid:dict,
    pts:_pd_DataFrame,
    y:str='lat',
    x:str='lon',
    row_name:str='id_y',
    col_name:str='id_x',
    silent:bool=False,
) -> _pd_DataFrame:
    """
    # TODO Move to class and Properly describe.
    # TODO it modifies pts AND grid?
    Modifies input pandas.DataFrame grid and pts: 
    sorts by 1) y coordinate and 2) by x coordinate

    Args:
    <y>
    
    Returns:
    gridcell_id_name: name to be appended in pts to indicate gridcell. If False then information will not be stored in pts 
    """
    if not silent:
        print(
            'Aggregate Data from '+str(len(pts))+' points'+
            ' into '+str(len(grid.row_ids))+'x'+str(len(grid.col_ids))+
            '='+str(grid.n_cells)+' cells.' 
        )
    # to do change to cut
    # for each row select relevant points, then refine selection with columns to obtain cells
    pts[row_name] = ((pts[y]-grid.total_bounds.ymin) // grid.spacing).astype(int)
    pts[col_name] = ((pts[x]-grid.total_bounds.xmin) // grid.spacing).astype(int)
    
    # nest_height = 4
    # for i in range(1,nest_height+1):
    #     for j in range(i*2):
    #         pts[find_column_name(row_name+'_'+str(i)+'_'+str(j), pts.columns)] = ((pts[y]-grid.total_bounds.ymin+j*grid.spacing) // (i*grid.spacing)).astype(int)
    #         pts[find_column_name(col_name+'_'+str(i)+'_'+str(j), pts.columns)] = ((pts[x]-grid.total_bounds.xmin+j*grid.spacing) // (i*grid.spacing)).astype(int)
    
    return pts[[row_name, col_name]]

def translate_subcell_row_col_to_value(row_nr, col_nr, lvl, nest_depth):
    """TODO create function that that translates subcell row col to value"""
    subcell_value = 0
    for i in range(1,nest_depth+1):
            n = 2**(i)
            subcell_mult = 2**((nest_depth-i)*2)
            print("nest_depth", nest_depth, "i", i, "subcell_mult",subcell_mult, "n", n)
            print([int(x//(1/n)%n%2) for x in _np_linspace(0,1,10)])
            # TODO this is not yet ready.
            subcell_value += (col_nr//(1/n)%n%2 * subcell_mult  + row_nr//(1/n)%n%2 * 2 * subcell_mult).astype(int)
        #
    return subcell_value

@time_func_perf
def aggregate_point_data_to_cells(
    grid:dict,
    pts:_pd_DataFrame,
    c:list=['employment'],
    y:str='lat',
    x:str='lon',
    row_name:str='id_y',
    col_name:str='id_x',
    nest_depth:int=5,
    nest_height:int=5,
    silent:bool=False,
) -> _pd_DataFrame:
    """
    TODO
    """
    # what is points data initally sorted by
    # aggregate cells to super cells to save lookup time 
    aggregate_level = 0

    # sort by row, then by col = resulting in cell wise sorting
    # then sort for quadrants 
    cols_for_sort = [row_name, col_name]
    subcell_nr = find_column_name('sc_nr', existing_columns=pts.columns)

    
    if nest_depth > 0:
        # offsets normalized to 0-1
        offset_x = 0.5 + ((pts[x]-grid.total_bounds.xmin)%grid.spacing - grid.spacing/2) / grid.spacing
        offset_y = 0.5 + ((pts[y]-grid.total_bounds.ymin)%grid.spacing - grid.spacing/2) / grid.spacing
        pts[subcell_nr] = 0

        # loop through nest levels starting from the broadest/most aggregate end with smallest/most narrow
        for i in range(1,nest_depth+1):
            n = 2**(i)
            subcell_mult = 2**((nest_depth-i)*2)
            # print("nest_depth", nest_depth, "i", i, "subcell_mult",subcell_mult, "n", n)
            # print([int(x//(1/n)%n%2) for x in _np_linspace(0,1,10)])
            pts[subcell_nr] += (offset_x//(1/n)%n%2 * subcell_mult  + offset_y//(1/n)%n%2 * 2 * subcell_mult).astype(int)
        #
        cols_for_sort.append(subcell_nr)
    #
    
    pts.sort_values(cols_for_sort,inplace=True)
    
    # extract variables from dataframe for faster access speed
    row_ids = pts[row_name].unique().tolist()
    col_ids = pts[col_name].unique().tolist()
    pts_rows = pts[row_name].tolist()
    pts_cols = pts[col_name].values
    pts_cols_list = pts[col_name].tolist()
    pts_vals = pts[c].values #if type(c)==list and len(c)>1 else pts[c[0]].values
    pts_vals_xy = pts[c+[x,y]].values
    pts_ids = pts.index.values #tuple([int(x) for x in pts.index])
    pts_subcell_nrs = pts[subcell_nr].values if nest_depth > 0 else _np_zeros(len(pts_cols))
    n_pts, col_max = len(pts), col_ids[-1]

    # save position of first point in row
    pos, row_id_indexes = 0, []
    for i, row_id in enumerate(row_ids):
        row_id_indexes.append(pos+next((idx for idx, val in enumerate(pts_rows[pos:]) if val==row_id),None))   
        pos = row_id_indexes[i]+1 
    
    # output dictionarys
    id_to_sums, id_to_pt_ids, id_to_vals_xy = {}, {}, {}
    id_to_pt_ids_by_lvl = {}#{i: {} for i in range(1, nest_depth+1)}
    id_to_sums_by_lvl = {}#{i: {} for i in range(1, nest_depth+1)}
    id_to_vals_xy_by_lvl = {}#{i: {} for i in range(1, nest_depth+1)}
    
    # TODO: remove next line after testing
    tot = {"0":sum(pts_vals), **{i:0 for i in range(1, nest_depth+1) }}

    def nest_next_lvl(
            pos_min:int,
            pos_max:int,
            subcell_val:int=0,
            lvl:int=1,
            row:int=0,
            col:int=0,row_col=(0,0)
        ):
        """
        function that recursively splits cell into quadrants until nest_depth is reached: 
        2 subcell rows and 2 subcell columns and store the result of each non empty quadrant in dictionary
        """

        subcell_mult = 2**((nest_depth-lvl)*2)
        
        for cur_quad_nr in range(4):
            # 
            if pos_min >= pos_max:
                break
            # if there are no points in quadrant, continue with next
            if (pts_subcell_nrs[pos_min]-subcell_val) // subcell_mult > cur_quad_nr:
                continue
            # otherwise there are points in quadrant
            # find first point not in quadrant, if none include all remaining points 
            pos_next = next((i+pos_min for i, subcell_nr in enumerate(pts_subcell_nrs[pos_min:pos_max]) if (subcell_nr-subcell_val)//subcell_mult>cur_quad_nr), pos_max)
            # TODO: remove next line after testing
            
            next_subcell_val = subcell_val + cur_quad_nr * subcell_mult
            row_c = row + cur_quad_nr//2 * 1/(2**lvl)
            col_c = col + cur_quad_nr%2 * 1/(2**lvl)
            row_col_centroids = (row_c+2**-(lvl+1),col_c+2**-(lvl+1))
            id_to_sums_by_lvl[(lvl,row_col_centroids)] = pts_vals[pos_min:pos_next].sum(axis=0)
            id_to_pt_ids_by_lvl[(lvl,row_col_centroids)] = pts_ids[pos_min:pos_next]
            id_to_vals_xy_by_lvl[(lvl,row_col_centroids)] = pts_vals_xy[pos_min:pos_next]
            if lvl+1 <= nest_depth:
                nest_next_lvl(
                    pos_min=pos_min,
                    pos_max=pos_next,
                    subcell_val=next_subcell_val,
                    lvl=lvl+1, row=row_c, col=col_c, row_col=row_col,
                )
            pos_min = pos_next
        #
    #

    # loop over rows
    for cur_row, cur_col_i, next_row_i, cur_col in zip(row_ids, row_id_indexes, row_id_indexes[1:]+[n_pts], pts_cols[row_id_indexes]):
        # somwhere np.int64 was introduced thus reapply int() type. Could be done more elagantly if source of np.int64 is identified
        cur_row, cur_col_i, next_row_i, cur_col = int(cur_row), int(cur_col_i), int(next_row_i), int(cur_col)
        # find next column to the right until all points from row an included in a column of the row
        while True:
            # for points of the same row that are to the right of current column, find first and return its position and column 
            next_col_i, next_col = next(((i+cur_col_i,c) for i,c in enumerate(pts_cols_list[cur_col_i:next_row_i]) if cur_col<c),(next_row_i,col_max+1)) 
            # if there are no points to the right of the column break to continue with the next row
            if cur_col_i >= next_col_i:
                break
            
            # for points within resulting cell:
            # aggregate data
            id_to_sums[(cur_row,cur_col)] = pts_vals[cur_col_i:next_col_i].sum(axis=0)
            # store point ids
            id_to_pt_ids[(cur_row,cur_col)] = pts_ids[cur_col_i:next_col_i]
            # store point coords
            id_to_vals_xy[(cur_row,cur_col)] = pts_vals_xy[cur_col_i:next_col_i]

            id_to_sums_by_lvl[(0,(cur_row,cur_col))] = pts_vals[cur_col_i:next_col_i].sum(axis=0)
            # TODO the next two are not needed as no cell will be counted as overlapped 
            # if nest_depth > 0:
            id_to_pt_ids_by_lvl[(0,(cur_row,cur_col))] = pts_ids[cur_col_i:next_col_i]
            id_to_vals_xy_by_lvl[(0,(cur_row,cur_col))] = pts_vals_xy[cur_col_i:next_col_i]
            # now create subcells and aggregate data within it
            if nest_depth > 0:
                nest_next_lvl(
                    pos_min=cur_col_i,
                    pos_max=next_col_i,
                    subcell_val=0,
                    lvl=1, row=cur_row-.5,col=cur_col-.5,row_col=(cur_row,cur_col)
                )
            cur_col_i, cur_col = next_col_i, next_col
        #
    #
    id_to_n_pts_by_lvl={}
    ref_lvl = grid.ref_lvl
    if ref_lvl < 0:
        ref_lvl_factor = 2**abs(grid.ref_lvl)
        id_to_pt_ids_by_ref_lvl = {}
        id_to_n_pts_by_ref_lvl = {}
        id_to_sums_by_ref_lvl = {}
        id_to_pt_vals_xy_by_ref_lvl = {}
        for ((row,col), pt_ids), sums, pt_vals_xy in zip(id_to_pt_ids.items(), id_to_sums.values(), id_to_vals_xy.values()):
            ref_cell = (row%ref_lvl_factor, col%ref_lvl_factor)
            if ref_cell in id_to_pt_ids_by_ref_lvl:
                id_to_sums_by_lvl[(ref_lvl,ref_cell)] += sums
                id_to_pt_ids_by_lvl[(ref_lvl, ref_cell)] += list(pt_ids)
                id_to_vals_xy_by_lvl[((ref_lvl, ref_cell))] += list(pt_vals_xy)
                id_to_n_pts_by_lvl[(ref_lvl,ref_cell)] += len(pt_ids)

            else:
                # ensure to create a copy of each data type
                id_to_sums_by_lvl[(ref_lvl,ref_cell)] = sums*1
                id_to_pt_ids_by_lvl[(ref_lvl,ref_cell)] = list(pt_ids)
                id_to_n_pts_by_lvl[(ref_lvl,ref_cell)] = len(pt_ids)
                id_to_vals_xy_by_lvl[(ref_lvl,ref_cell)] = list(pt_vals_xy)
        
        grid.id_to_sums_by_lvl[grid.ref_lvl] = id_to_sums_by_ref_lvl
        grid.id_to_pt_ids_by_lvl[grid.ref_lvl] = {k:_np_array(v,int) for k,v in id_to_pt_ids_by_ref_lvl.items()}
        grid.id_to_pt_vals_xy_by_lvl[grid.ref_lvl] = {k:_np_array(v,float) for k,v in id_to_pt_vals_xy_by_ref_lvl.items()}
    
    # TODO dictionaries that are no longer needed
    grid.id_to_sums = id_to_sums
    grid.id_to_pt_ids = id_to_pt_ids
    grid.id_to_vals_xy = id_to_vals_xy

    grid.id_to_pt_ids_by_lvl = id_to_pt_ids_by_lvl
    grid.id_to_sums_by_lvl = id_to_sums_by_lvl
    grid.id_to_vals_xy_by_lvl = id_to_vals_xy_by_lvl
    subsubtypes_sparse_grid = set()
    for row_col in set(id_to_pt_ids):
        for x in row_col:
            subsubtypes_sparse_grid.add(type(x))
    # TODO drop column 
    # if nest_depth > 0:
    #     pts.drop(columns=[subcell_nr], inplace=True)
    
    if not silent:
        print('Total(s) for grid cells:', _np_array([s for s in grid.id_to_sums.values()]).sum(axis=0), 'Total(s) for points:', pts[c].values.sum(axis=0))
    #
    return 
#



import sys
from typing import Any

import numpy as np
import numpy.typing as npt

if sys.version_info >= (3, 11):
    from typing import assert_type
else:
    from typing_extensions import assert_type

AR_LIKE_b: list[bool]
AR_LIKE_u: list[np.uint32]
AR_LIKE_i: list[int]
AR_LIKE_f: list[float]
AR_LIKE_O: list[np.object_]

AR_U: npt.NDArray[np.str_]

assert_type(np.fix(AR_LIKE_b), npt.NDArray[np.floating[Any]])
assert_type(np.fix(AR_LIKE_u), npt.NDArray[np.floating[Any]])
assert_type(np.fix(AR_LIKE_i), npt.NDArray[np.floating[Any]])
assert_type(np.fix(AR_LIKE_f), npt.NDArray[np.floating[Any]])
assert_type(np.fix(AR_LIKE_O), npt.NDArray[np.object_])
assert_type(np.fix(AR_LIKE_f, out=AR_U), npt.NDArray[np.str_])

assert_type(np.isposinf(AR_LIKE_b), npt.NDArray[np.bool])
assert_type(np.isposinf(AR_LIKE_u), npt.NDArray[np.bool])
assert_type(np.isposinf(AR_LIKE_i), npt.NDArray[np.bool])
assert_type(np.isposinf(AR_LIKE_f), npt.NDArray[np.bool])
assert_type(np.isposinf(AR_LIKE_f, out=AR_U), npt.NDArray[np.str_])

assert_type(np.isneginf(AR_LIKE_b), npt.NDArray[np.bool])
assert_type(np.isneginf(AR_LIKE_u), npt.NDArray[np.bool])
assert_type(np.isneginf(AR_LIKE_i), npt.NDArray[np.bool])
assert_type(np.isneginf(AR_LIKE_f), npt.NDArray[np.bool])
assert_type(np.isneginf(AR_LIKE_f, out=AR_U), npt.NDArray[np.str_])



def __getattr__(attr_name):
    import warnings
    from numpy.fft import _helper
    ret = getattr(_helper, attr_name, None)
    if ret is None:
        raise AttributeError(
            f"module 'numpy.fft.helper' has no attribute {attr_name}")
    warnings.warn(
        "The numpy.fft.helper has been made private and renamed to "
        "numpy.fft._helper. All four functions exported by it (i.e. fftshift, "
        "ifftshift, fftfreq, rfftfreq) are available from numpy.fft. "
        f"Please use numpy.fft.{attr_name} instead.",
        DeprecationWarning,
        stacklevel=3
    )
    return ret



from __future__ import annotations

import sys
from collections.abc import Collection, Callable, Sequence
from typing import Any, Protocol, Union, TypeVar, runtime_checkable

import numpy as np
from numpy import (
    ndarray,
    dtype,
    generic,
    unsignedinteger,
    integer,
    floating,
    complexfloating,
    number,
    timedelta64,
    datetime64,
    object_,
    void,
    str_,
    bytes_,
)
from ._nested_sequence import _NestedSequence

_T = TypeVar("_T")
_ScalarType = TypeVar("_ScalarType", bound=generic)
_ScalarType_co = TypeVar("_ScalarType_co", bound=generic, covariant=True)
_DType = TypeVar("_DType", bound=dtype[Any])
_DType_co = TypeVar("_DType_co", covariant=True, bound=dtype[Any])

NDArray = ndarray[Any, dtype[_ScalarType_co]]

# The `_SupportsArray` protocol only cares about the default dtype
# (i.e. `dtype=None` or no `dtype` parameter at all) of the to-be returned
# array.
# Concrete implementations of the protocol are responsible for adding
# any and all remaining overloads
@runtime_checkable
class _SupportsArray(Protocol[_DType_co]):
    def __array__(self) -> ndarray[Any, _DType_co]: ...


@runtime_checkable
class _SupportsArrayFunc(Protocol):
    """A protocol class representing `~class.__array_function__`."""
    def __array_function__(
        self,
        func: Callable[..., Any],
        types: Collection[type[Any]],
        args: tuple[Any, ...],
        kwargs: dict[str, Any],
    ) -> object: ...


# TODO: Wait until mypy supports recursive objects in combination with typevars
_FiniteNestedSequence = Union[
    _T,
    Sequence[_T],
    Sequence[Sequence[_T]],
    Sequence[Sequence[Sequence[_T]]],
    Sequence[Sequence[Sequence[Sequence[_T]]]],
]

# A subset of `npt.ArrayLike` that can be parametrized w.r.t. `np.generic`
_ArrayLike = Union[
    _SupportsArray[dtype[_ScalarType]],
    _NestedSequence[_SupportsArray[dtype[_ScalarType]]],
]

# A union representing array-like objects; consists of two typevars:
# One representing types that can be parametrized w.r.t. `np.dtype`
# and another one for the rest
_DualArrayLike = Union[
    _SupportsArray[_DType],
    _NestedSequence[_SupportsArray[_DType]],
    _T,
    _NestedSequence[_T],
]

if sys.version_info >= (3, 12):
    from collections.abc import Buffer

    ArrayLike = Buffer | _DualArrayLike[
        dtype[Any],
        Union[bool, int, float, complex, str, bytes],
    ]
else:
    ArrayLike = _DualArrayLike[
        dtype[Any],
        Union[bool, int, float, complex, str, bytes],
    ]

# `ArrayLike<X>_co`: array-like objects that can be coerced into `X`
# given the casting rules `same_kind`
_ArrayLikeBool_co = _DualArrayLike[
    dtype[np.bool],
    bool,
]
_ArrayLikeUInt_co = _DualArrayLike[
    dtype[Union[np.bool, unsignedinteger[Any]]],
    bool,
]
_ArrayLikeInt_co = _DualArrayLike[
    dtype[Union[np.bool, integer[Any]]],
    Union[bool, int],
]
_ArrayLikeFloat_co = _DualArrayLike[
    dtype[Union[np.bool, integer[Any], floating[Any]]],
    Union[bool, int, float],
]
_ArrayLikeComplex_co = _DualArrayLike[
    dtype[Union[
        np.bool,
        integer[Any],
        floating[Any],
        complexfloating[Any, Any],
    ]],
    Union[bool, int, float, complex],
]
_ArrayLikeNumber_co = _DualArrayLike[
    dtype[Union[np.bool, number[Any]]],
    Union[bool, int, float, complex],
]
_ArrayLikeTD64_co = _DualArrayLike[
    dtype[Union[np.bool, integer[Any], timedelta64]],
    Union[bool, int],
]
_ArrayLikeDT64_co = Union[
    _SupportsArray[dtype[datetime64]],
    _NestedSequence[_SupportsArray[dtype[datetime64]]],
]
_ArrayLikeObject_co = Union[
    _SupportsArray[dtype[object_]],
    _NestedSequence[_SupportsArray[dtype[object_]]],
]

_ArrayLikeVoid_co = Union[
    _SupportsArray[dtype[void]],
    _NestedSequence[_SupportsArray[dtype[void]]],
]
_ArrayLikeStr_co = _DualArrayLike[
    dtype[str_],
    str,
]
_ArrayLikeBytes_co = _DualArrayLike[
    dtype[bytes_],
    bytes,
]

_ArrayLikeInt = _DualArrayLike[
    dtype[integer[Any]],
    int,
]

# Extra ArrayLike type so that pyright can deal with NDArray[Any]
# Used as the first overload, should only match NDArray[Any],
# not any actual types.
# https://github.com/numpy/numpy/pull/22193
class _UnknownType:
    ...


_ArrayLikeUnknown = _DualArrayLike[
    dtype[_UnknownType],
    _UnknownType,
]
